from abc import ABC, abstractmethod
from typing import Any

from bepatient.waiter_src.checkers.checker import Checker
from bepatient.waiter_src.conditions_manager import ConditionsManager
from bepatient.waiter_src.exceptions import ExecutorIsNotReady


class Executor(ABC):
    """An abstract base class for defining an executor that can be waited for."""

    def __init__(self):
        self.conditions_manager = ConditionsManager()
        self._failed_checkers: list[Checker] = []
        self._result: Any = None
        self._input: str | None = None

    def add_exception_condition(self, checker: Checker):
        """Adds checker function to the condition's manager. If the checker condition
        is not met, it raises ExceptionConditionNotMet."""
        self.conditions_manager.exception_conditions.append(checker)
        return self

    def add_pre_condition(self, checker: Checker):
        """Adds checker function to the list of pre_conditions - conditions that will
        be checked before the main ones and after those that may result in an error.
        For example, when we want to check the response status code before verifying
        the content of the body."""
        self.conditions_manager.pre_conditions.append(checker)
        return self

    def add_main_condition(self, checker: Checker):
        """Adds checker function to the list of main_conditions in ConditionsManager.
        These are the main conditions to be met. That which we really care about
        checking the most."""
        self.conditions_manager.main_conditions.append(checker)
        return self

    @abstractmethod
    def is_condition_met(self) -> bool:
        """Check whether the condition has been met.

        Returns:
            bool: True if the condition has been met, False otherwise."""

    def get_result(self) -> Any:
        """Returns the result of performed actions."""
        if self._result is not None:
            return self._result
        raise ExecutorIsNotReady()

    def error_message(self) -> str:
        """Return a detailed error message if the condition has not been met."""
        if self._result is not None and len(self._failed_checkers) > 0:
            checkers = ", ".join([str(checker) for checker in self._failed_checkers])
            return (
                "The condition has not been met!"
                f" | Failed checkers: ({checkers})"
                f" | {self._input}"
            )
        if self._result is not None and len(self._failed_checkers) == 0:
            return "All conditions have been met."
        raise ExecutorIsNotReady()



import logging
from typing import Any, Literal

from bepatient.waiter_src.checkers.checker import Checker
from bepatient.waiter_src.exceptions import ExceptionConditionNotMet, WaiterIsNotReady

log = logging.getLogger(__name__)
CONDITION_LEVEL = Literal["exception", "pre", "main"]  # pylint: disable=invalid-name


class ConditionsManager:
    """Manages and evaluates conditions grouped into three levels:
    - exception_conditions - conditions that trigger an exception if not met,
    - precondition_conditions - conditions that are checked prior to the main
        conditions,
    - main_conditions - core conditions, those are checked last."""

    def __init__(self):
        self.exception_conditions = []
        self.pre_conditions = []
        self.main_conditions = []

    def check_all(self, result: Any, check_uuid: str) -> list[Checker]:
        """Simply checks all defined conditions."""
        if not any(
            (self.exception_conditions, self.pre_conditions, self.main_conditions)
        ):
            raise WaiterIsNotReady("No conditions defined")
        if not self.main_conditions:
            log.info("No main conditions available")

        if self.exception_conditions:
            failed_checkers = [
                checker
                for checker in self.exception_conditions
                if not checker.check(result, check_uuid)
            ]
            if failed_checkers:
                checkers = ", ".join((str(checker) for checker in failed_checkers))
                raise ExceptionConditionNotMet(f"Failed checkers: {checkers}")

        if self.pre_conditions:
            failed_checkers = [
                checker
                for checker in self.pre_conditions
                if not checker.check(result, check_uuid)
            ]
            if failed_checkers:
                return failed_checkers
        return [
            checker
            for checker in self.main_conditions
            if not checker.check(result, check_uuid)
        ]



class BePatientException(Exception):
    pass


class ExecutorIsNotReady(BePatientException):
    """We raise it when user want to use `get_result` or `error_message` methods from
    the `Executor` before checking if the condition has been met."""

    def __init__(self, message: str = "The condition has not yet been checked."):
        super().__init__(message)


class WaiterIsNotReady(BePatientException):
    """Not all required attributes have been set."""


class WaiterConditionWasNotMet(BePatientException):
    """Conditions were not met within the specified number of attempts."""


class ExceptionConditionNotMet(BePatientException):
    """One of the conditions causing the wait for the result to end has not been met."""
